/*
  This file is part of TALER
  Copyright (C) 2014-2026 Taler Systems SA

  TALER is free software; you can redistribute it and/or modify it under the
  terms of the GNU General Public License as published by the Free Software
  Foundation; either version 3, or (at your option) any later version.

  TALER is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
  A PARTICULAR PURPOSE.  See the GNU General Public License for more details.

  You should have received a copy of the GNU General Public License along with
  TALER; see the file COPYING.  If not, see
  <http://www.gnu.org/licenses/>
*/
/**
 * @file lib/exchange_api_get-deposits-H_WIRE-MERCHANT_PUB-H_CONTRACT_TERMS-COIN_PUB.c
 * @brief Implementation of the GET /deposits/... request
 * @author Christian Grothoff
 */
#include <jansson.h>
#include <microhttpd.h> /* just for HTTP status codes */
#include <gnunet/gnunet_util_lib.h>
#include <gnunet/gnunet_json_lib.h>
#include <gnunet/gnunet_curl_lib.h>
#include "taler/taler_json_lib.h"
#include "exchange_api_handle.h"
#include "taler/taler_signatures.h"
#include "exchange_api_curl_defaults.h"


/**
 * @brief A GET /deposits/... Handle
 */
struct TALER_EXCHANGE_GetDepositsHandle
{

  /**
   * Base URL of the exchange.
   */
  char *base_url;

  /**
   * The url for this request.
   */
  char *url;

  /**
   * The keys of this request handle will use.
   */
  struct TALER_EXCHANGE_Keys *keys;

  /**
   * Handle for the request.
   */
  struct GNUNET_CURL_Job *job;

  /**
   * Function to call with the result.
   */
  TALER_EXCHANGE_GetDepositsCallback cb;

  /**
   * Closure for @e cb.
   */
  TALER_EXCHANGE_GET_DEPOSITS_RESULT_CLOSURE *cb_cls;

  /**
   * CURL context to use.
   */
  struct GNUNET_CURL_Context *ctx;

  /**
   * Private key of the merchant.
   */
  struct TALER_MerchantPrivateKeyP merchant_priv;

  /**
   * Hash over the wiring information of the merchant.
   */
  struct TALER_MerchantWireHashP h_wire;

  /**
   * Hash over the contract for which this deposit is made.
   */
  struct TALER_PrivateContractHashP h_contract_terms;

  /**
   * The coin's public key.
   */
  struct TALER_CoinSpendPublicKeyP coin_pub;

  /**
   * Options set for this request.
   */
  struct
  {
    /**
     * How long to wait for the wire transfer, enabling long polling.
     * Default: zero (no long polling).
     */
    struct GNUNET_TIME_Relative timeout;

    /**
     * Long-poll target type.  Defaults to #TALER_DGLPT_NONE (0).
     */
    enum TALER_DepositGetLongPollTarget lpt;
  } options;

};


/**
 * Function called when we're done processing the
 * HTTP GET /deposits/... request.
 *
 * @param cls the `struct TALER_EXCHANGE_GetDepositsHandle`
 * @param response_code HTTP response code, 0 on error
 * @param response parsed JSON result, NULL on error
 */
static void
handle_deposit_wtid_finished (void *cls,
                              long response_code,
                              const void *response)
{
  struct TALER_EXCHANGE_GetDepositsHandle *gdh = cls;
  const json_t *j = response;
  struct TALER_EXCHANGE_GetDepositsResponse dr = {
    .hr.reply = j,
    .hr.http_status = (unsigned int) response_code
  };

  gdh->job = NULL;
  switch (response_code)
  {
  case 0:
    dr.hr.ec = TALER_EC_GENERIC_INVALID_RESPONSE;
    break;
  case MHD_HTTP_OK:
    {
      struct GNUNET_JSON_Specification spec[] = {
        GNUNET_JSON_spec_fixed_auto (
          "wtid",
          &dr.details.ok.wtid),
        GNUNET_JSON_spec_timestamp (
          "execution_time",
          &dr.details.ok.execution_time),
        TALER_JSON_spec_amount_any (
          "coin_contribution",
          &dr.details.ok.coin_contribution),
        GNUNET_JSON_spec_fixed_auto (
          "exchange_sig",
          &dr.details.ok.exchange_sig),
        GNUNET_JSON_spec_fixed_auto (
          "exchange_pub",
          &dr.details.ok.exchange_pub),
        GNUNET_JSON_spec_end ()
      };

      GNUNET_assert (NULL != gdh->keys);
      if (GNUNET_OK !=
          GNUNET_JSON_parse (j,
                             spec,
                             NULL, NULL))
      {
        GNUNET_break_op (0);
        dr.hr.http_status = 0;
        dr.hr.ec = TALER_EC_GENERIC_REPLY_MALFORMED;
        break;
      }
      if (GNUNET_OK !=
          TALER_EXCHANGE_test_signing_key (gdh->keys,
                                           &dr.details.ok.exchange_pub))
      {
        GNUNET_break_op (0);
        dr.hr.http_status = 0;
        dr.hr.ec = TALER_EC_EXCHANGE_DEPOSITS_GET_INVALID_SIGNATURE_BY_EXCHANGE;
        break;
      }
      if (GNUNET_OK !=
          TALER_exchange_online_confirm_wire_verify (
            &gdh->h_wire,
            &gdh->h_contract_terms,
            &dr.details.ok.wtid,
            &gdh->coin_pub,
            dr.details.ok.execution_time,
            &dr.details.ok.coin_contribution,
            &dr.details.ok.exchange_pub,
            &dr.details.ok.exchange_sig))
      {
        GNUNET_break_op (0);
        dr.hr.http_status = 0;
        dr.hr.ec = TALER_EC_EXCHANGE_DEPOSITS_GET_INVALID_SIGNATURE_BY_EXCHANGE;
        break;
      }
      gdh->cb (gdh->cb_cls,
               &dr);
      gdh->cb = NULL;
      TALER_EXCHANGE_get_deposits_cancel (gdh);
      return;
    }
  case MHD_HTTP_ACCEPTED:
    {
      /* Transaction known, but not executed yet */
      bool no_legi = false;
      struct GNUNET_JSON_Specification spec[] = {
        GNUNET_JSON_spec_timestamp (
          "execution_time",
          &dr.details.accepted.execution_time),
        GNUNET_JSON_spec_mark_optional (
          GNUNET_JSON_spec_fixed_auto (
            "account_pub",
            &dr.details.accepted.account_pub),
          NULL),
        GNUNET_JSON_spec_mark_optional (
          GNUNET_JSON_spec_uint64 (
            "requirement_row",
            &dr.details.accepted.requirement_row),
          &no_legi),
        GNUNET_JSON_spec_bool (
          "kyc_ok",
          &dr.details.accepted.kyc_ok),
        GNUNET_JSON_spec_end ()
      };

      if (GNUNET_OK !=
          GNUNET_JSON_parse (j,
                             spec,
                             NULL, NULL))
      {
        GNUNET_break_op (0);
        dr.hr.http_status = 0;
        dr.hr.ec = TALER_EC_GENERIC_REPLY_MALFORMED;
        break;
      }
      if (no_legi)
        dr.details.accepted.requirement_row = 0;
      gdh->cb (gdh->cb_cls,
               &dr);
      gdh->cb = NULL;
      TALER_EXCHANGE_get_deposits_cancel (gdh);
      return;
    }
  case MHD_HTTP_BAD_REQUEST:
    dr.hr.ec = TALER_JSON_get_error_code (j);
    dr.hr.hint = TALER_JSON_get_error_hint (j);
    /* This should never happen, either us or the exchange is buggy
       (or API version conflict); just pass JSON reply to the application */
    break;
  case MHD_HTTP_FORBIDDEN:
    dr.hr.ec = TALER_JSON_get_error_code (j);
    dr.hr.hint = TALER_JSON_get_error_hint (j);
    /* Nothing really to verify, exchange says one of the signatures is
       invalid; as we checked them, this should never happen, we
       should pass the JSON reply to the application */
    break;
  case MHD_HTTP_NOT_FOUND:
    dr.hr.ec = TALER_JSON_get_error_code (j);
    dr.hr.hint = TALER_JSON_get_error_hint (j);
    /* Exchange does not know about transaction;
       we should pass the reply to the application */
    break;
  case MHD_HTTP_INTERNAL_SERVER_ERROR:
    dr.hr.ec = TALER_JSON_get_error_code (j);
    dr.hr.hint = TALER_JSON_get_error_hint (j);
    /* Server had an internal issue; we should retry, but this API
       leaves this to the application */
    break;
  default:
    /* unexpected response code */
    dr.hr.ec = TALER_JSON_get_error_code (j);
    dr.hr.hint = TALER_JSON_get_error_hint (j);
    GNUNET_log (GNUNET_ERROR_TYPE_ERROR,
                "Unexpected response code %u/%d for exchange GET deposits\n",
                (unsigned int) response_code,
                (int) dr.hr.ec);
    GNUNET_break_op (0);
    break;
  }
  if (NULL != gdh->cb)
    gdh->cb (gdh->cb_cls,
             &dr);
  TALER_EXCHANGE_get_deposits_cancel (gdh);
}


struct TALER_EXCHANGE_GetDepositsHandle *
TALER_EXCHANGE_get_deposits_create (
  struct GNUNET_CURL_Context *ctx,
  const char *url,
  struct TALER_EXCHANGE_Keys *keys,
  const struct TALER_MerchantPrivateKeyP *merchant_priv,
  const struct TALER_MerchantWireHashP *h_wire,
  const struct TALER_PrivateContractHashP *h_contract_terms,
  const struct TALER_CoinSpendPublicKeyP *coin_pub)
{
  struct TALER_EXCHANGE_GetDepositsHandle *gdh;
  struct TALER_MerchantPublicKeyP merchant;

  gdh = GNUNET_new (struct TALER_EXCHANGE_GetDepositsHandle);
  gdh->ctx = ctx;
  gdh->base_url = GNUNET_strdup (url);
  gdh->keys = TALER_EXCHANGE_keys_incref (keys);
  gdh->merchant_priv = *merchant_priv;
  gdh->h_wire = *h_wire;
  gdh->h_contract_terms = *h_contract_terms;
  GNUNET_CRYPTO_eddsa_key_get_public (&merchant_priv->eddsa_priv,
                                      &merchant.eddsa_pub);
  gdh->coin_pub = *coin_pub;
  return gdh;
}


enum GNUNET_GenericReturnValue
TALER_EXCHANGE_get_deposits_set_options_ (
  struct TALER_EXCHANGE_GetDepositsHandle *gdh,
  unsigned int num_options,
  const struct TALER_EXCHANGE_GetDepositsOptionValue *options)
{
  for (unsigned int i = 0; i < num_options; i++)
  {
    const struct TALER_EXCHANGE_GetDepositsOptionValue *opt = &options[i];

    switch (opt->option)
    {
    case TALER_EXCHANGE_GET_DEPOSITS_OPTION_END:
      return GNUNET_OK;
    case TALER_EXCHANGE_GET_DEPOSITS_OPTION_TIMEOUT:
      gdh->options.timeout = opt->details.timeout;
      break;
    case TALER_EXCHANGE_GET_DEPOSITS_OPTION_LONG_POLL_TARGET:
      gdh->options.lpt = opt->details.lpt;
      break;
    }
  }
  return GNUNET_OK;
}


enum TALER_ErrorCode
TALER_EXCHANGE_get_deposits_start (
  struct TALER_EXCHANGE_GetDepositsHandle *gdh,
  TALER_EXCHANGE_GetDepositsCallback cb,
  TALER_EXCHANGE_GET_DEPOSITS_RESULT_CLOSURE *cb_cls)
{
  struct TALER_MerchantPublicKeyP merchant;
  struct TALER_MerchantSignatureP merchant_sig;
  unsigned int tms;
  CURL *eh;

  if (NULL != gdh->job)
  {
    GNUNET_break (0);
    return TALER_EC_GENERIC_INTERNAL_INVARIANT_FAILURE;
  }
  gdh->cb = cb;
  gdh->cb_cls = cb_cls;
  tms = (unsigned int) gdh->options.timeout.rel_value_us
        / GNUNET_TIME_UNIT_MILLISECONDS.rel_value_us;
  GNUNET_CRYPTO_eddsa_key_get_public (&gdh->merchant_priv.eddsa_priv,
                                      &merchant.eddsa_pub);
  TALER_merchant_deposit_sign (&gdh->h_contract_terms,
                               &gdh->h_wire,
                               &gdh->coin_pub,
                               &gdh->merchant_priv,
                               &merchant_sig);
  {
    char cpub_str[sizeof (struct TALER_CoinSpendPublicKeyP) * 2];
    char mpub_str[sizeof (struct TALER_MerchantPublicKeyP) * 2];
    char msig_str[sizeof (struct TALER_MerchantSignatureP) * 2];
    char chash_str[sizeof (struct TALER_PrivateContractHashP) * 2];
    char whash_str[sizeof (struct TALER_MerchantWireHashP) * 2];
    char arg_str[sizeof (whash_str)
                 + sizeof (mpub_str)
                 + sizeof (chash_str)
                 + sizeof (cpub_str) + 48];
    char timeout_str[24];
    char lpt_str[12];
    char *end;

    end = GNUNET_STRINGS_data_to_string (&gdh->h_wire,
                                         sizeof (gdh->h_wire),
                                         whash_str,
                                         sizeof (whash_str));
    *end = '\0';
    end = GNUNET_STRINGS_data_to_string (&merchant,
                                         sizeof (merchant),
                                         mpub_str,
                                         sizeof (mpub_str));
    *end = '\0';
    end = GNUNET_STRINGS_data_to_string (&gdh->h_contract_terms,
                                         sizeof (gdh->h_contract_terms),
                                         chash_str,
                                         sizeof (chash_str));
    *end = '\0';
    end = GNUNET_STRINGS_data_to_string (&gdh->coin_pub,
                                         sizeof (gdh->coin_pub),
                                         cpub_str,
                                         sizeof (cpub_str));
    *end = '\0';
    end = GNUNET_STRINGS_data_to_string (&merchant_sig,
                                         sizeof (merchant_sig),
                                         msig_str,
                                         sizeof (msig_str));
    *end = '\0';
    if (0 == tms)
    {
      timeout_str[0] = '\0';
    }
    else
    {
      GNUNET_snprintf (
        timeout_str,
        sizeof (timeout_str),
        "%u",
        tms);
    }
    if (TALER_DGLPT_OK == gdh->options.lpt)
    {
      lpt_str[0] = '\0';
    }
    else
    {
      GNUNET_snprintf (
        lpt_str,
        sizeof (lpt_str),
        "%u",
        (unsigned int) gdh->options.lpt);
    }

    GNUNET_snprintf (arg_str,
                     sizeof (arg_str),
                     "deposits/%s/%s/%s/%s",
                     whash_str,
                     mpub_str,
                     chash_str,
                     cpub_str);
    gdh->url = TALER_url_join (
      gdh->base_url,
      arg_str,
      "merchant_sig", msig_str,
      "timeout_ms", (0 == tms) ? NULL : timeout_str,
      "lpt", (TALER_DGLPT_NONE == gdh->options.lpt) ? NULL : lpt_str,
      NULL);
  }
  if (NULL == gdh->url)
  {
    GNUNET_break (0);
    return TALER_EC_GENERIC_CONFIGURATION_INVALID;
  }
  eh = TALER_EXCHANGE_curl_easy_get_ (gdh->url);
  if (NULL == eh)
  {
    GNUNET_break (0);
    return TALER_EC_GENERIC_CONFIGURATION_INVALID;
  }
  if (0 != tms)
  {
    GNUNET_break (CURLE_OK ==
                  curl_easy_setopt (eh,
                                    CURLOPT_TIMEOUT_MS,
                                    (long) (tms + 100L)));
  }
  gdh->job = GNUNET_CURL_job_add (gdh->ctx,
                                  eh,
                                  &handle_deposit_wtid_finished,
                                  gdh);
  if (NULL == gdh->job)
    return TALER_EC_GENERIC_INTERNAL_INVARIANT_FAILURE;
  return TALER_EC_NONE;
}


void
TALER_EXCHANGE_get_deposits_cancel (
  struct TALER_EXCHANGE_GetDepositsHandle *gdh)
{
  if (NULL != gdh->job)
  {
    GNUNET_CURL_job_cancel (gdh->job);
    gdh->job = NULL;
  }
  GNUNET_free (gdh->url);
  GNUNET_free (gdh->base_url);
  TALER_EXCHANGE_keys_decref (gdh->keys);
  GNUNET_free (gdh);
}


/* end of exchange_api_get-deposits-H_WIRE-MERCHANT_PUB-H_CONTRACT_TERMS-COIN_PUB.c */
